from __future__ import division

import numpy as np

from .endclasses import endarray, tops


class EnergeticsBasicOld:
    """
    Energy functions based on SantaLucia's 2004 paper.  This is the "old"
    energetics class, which was originally used in Stickydesign.  It models
    DX-tile style sticky ends, but does not do so as specifically as
    the newer EnergeticsDAOE.

    You probably don't want to use this, unless you have preexisting code
    or systems that rely on this model:

    - for DX tiles, consider using EnergeticsDAOE
    - for toeholds, consider using EnergeticsBasic

    mismatchtype is one of 'max', 'loop', or 'dangle', specifying how to
    consider mismatches.  'max' is probably the best choice, but is slowest -
    it takes the maximum interaction of the 'loop' and 'dangle' options.
    """

    def __init__(self, mismatchtype='max'):
        import os
        try:
            import pkg_resources
            dsb = pkg_resources.resource_stream(__name__,
                                                os.path.join(
                                                    'params',
                                                    'dnastackingbig.csv'))
        except:
            try:
                this_dir, this_filename = os.path.split(__file__)
                dsb = open(
                    os.path.join(this_dir, "params", "dnastackingbig.csv"))
            except IOError:
                raise IOError("Error loading dnastackingbig.csv")
        self.nndG_full = -np.loadtxt(dsb, delimiter=',')
        dsb.close()
        self.initdG = 1.96
        self.nndG = self.nndG_full[np.arange(0, 16), 15 - np.arange(0, 16)]
        if mismatchtype == 'max':
            self.uniform = lambda x, y: \
                           np.maximum(self.uniform_loopmismatch(x, y),
                                      self.uniform_danglemismatch(x, y))
        elif mismatchtype == 'loop':
            self.uniform = self.uniform_loopmismatch
        elif mismatchtype == 'dangle':
            self.uniform = self.uniform_danglemismatch
        else:
            raise ValueError(
                "Mismatchtype {0} is not supported.".format(mismatchtype))

    def matching_uniform(self, seqs):
        return np.sum(self.nndG[tops(seqs)], axis=1) - self.initdG

    def uniform_loopmismatch(self, seqs1, seqs2):
        if seqs1.shape != seqs2.shape:
            if seqs1.ndim == 1:
                seqs1 = endarray(
                    np.repeat(np.array([seqs1]), seqs2.shape[0], 0),
                    seqs1.endtype)
            else:
                raise ValueError(
                    "Lengths of sequence arrays are not acceptable.")
        assert seqs1.endtype == seqs2.endtype
        endtype = seqs1.endtype

        endlen = seqs1.endlen
        plen = endlen - 1

        # TODO: replace this with cleaner code
        if endtype == 'DT':
            ps1 = seqs1[:, 1:-1] * 4 + seqs1[:, 2:]
            pa1 = seqs1[:, 0] * 4 + seqs1[:, 1]
            pac1 = (3 - seqs1[:, 0]) * 4 + seqs2[:, -1]
            ps2 = seqs2[:, ::-1][:, :-2] * 4 + seqs2[:, ::-1][:, 1:-1]
            pa2 = seqs2[:, 0] * 4 + seqs2[:, 1]
            pac2 = (3 - seqs2[:, 0]) * 4 + seqs1[:, -1]
        if endtype == 'TD':
            ps1 = seqs1[:, :-2] * 4 + seqs1[:, 1:-1]
            pa1 = seqs1[:, -2] * 4 + seqs1[:, -1]
            pac1 = seqs2[:, 0] * 4 + (3 - seqs1[:, -1])
            ps2 = seqs2[:, ::-1][:, 1:-1] * 4 + seqs2[:, ::-1][:, 2:]
            pa2 = seqs2[:, -2] * 4 + seqs2[:, -1]
            pac2 = (seqs1[:, 0]) * 4 + (3 - seqs2[:, -1])

        # Shift here is considering the first strand as fixed, and the second
        # one as shifting.  The shift is the offset of the bottom one in terms
        # of pair sequences (thus +2 and -1 instead of +1 and 0).
        en = np.zeros((ps1.shape[0], 2 * plen))
        for shift in range(-plen + 1, plen):
            en[:, plen + shift - 1] = np.sum(
                self.nndG_full[ps1[:, max(shift, 0):plen + shift],
                               ps2[:, max(-shift, 0):plen - shift]],
                axis=1)
        en[:, plen - 1] = (en[:, plen - 1] +
                           self.nndG_full[pa1, pac1] +
                           self.nndG_full[pa2, pac2])
        return np.amax(en, 1) - self.initdG

    def uniform_danglemismatch(self, seqs1, seqs2, fast=True):
        if seqs1.shape != seqs2.shape:
            if seqs1.ndim == 1:
                seqs1 = endarray(
                    np.repeat(np.array([seqs1]), seqs2.shape[0], 0),
                    seqs1.endtype)
            else:
                raise ValueError(
                    "Lengths of sequence arrays are not acceptable.")
        assert seqs1.endtype == seqs2.endtype
        endtype = seqs1.endtype
        s1 = tops(seqs1)
        s2 = tops(seqs2)
        l = s1.shape[1]
        s2r = np.fliplr(np.invert(s2) % 16)
        s2r = s2r // 4 + 4 * (s2r % 4)
        m = np.zeros((s1.shape[0], 2 * np.sum(np.arange(2, l + 1)) + l + 1))
        r = np.zeros(m.shape[0])
        z = 0
        if endtype == 'TD':
            s1c = s1[:, 0:-1]
            s2rc = s2r[:, 1:]
            s1l = np.hstack(((4 * (s2r[:, 0] // 4) + s1[:, 0] // 4).reshape(
                -1, 1), s1))
            s2rl = np.hstack(
                (s2r, (4 * (s2r[:, -1] % 4) + s1[:, -1] % 4).reshape(-1, 1)))
        elif endtype == 'DT':
            s1c = s1[:, 1:]
            s2rc = s2r[:, 0:-1]
            s2rl = np.hstack(((4 * (s1[:, 0] // 4) + s2r[:, 0] // 4).reshape(
                -1, 1), s2r))
            s1l = np.hstack(
                (s1, (4 * (s1[:, -1] % 4) + s2r[:, -1] % 4).reshape(-1, 1)))
        for o in range(1, l - 1):
            zn = l - 1 - o
            m[:, z:z + zn] = (
                s1c[:, :-o] == s2rc[:, o:]) * self.nndG[s1c[:, :-o]]
            z = z + zn + 2
            m[:, z:z + zn] = (
                s2rc[:, :-o] == s1c[:, o:]) * self.nndG[s2rc[:, :-o]]
            z = z + zn + 2
        m[:, z:z + l + 1] = (s1l == s2rl) * self.nndG[s1l]
        from ._stickyext import fastsub
        x = m
        fastsub(x, r)

        return r - self.initdG



import numpy as np
from math import ceil
from .endclasses import lton

import logging
LOGGER = logging.getLogger(__name__)


def endchooser(all_energetics,
               target_vals=None,
               templates=None,
               init_wigglefraction=1,
               next_wigglefraction=0.1,
               devmethod='dev'):
    """
    An endchooser generator that chooses ends while trying to optimize for
    multiple energy models simultaneously.

    Arguments:

    `all_energetics`: a list of energetics models to optimize for.

    `target_vals`: if provided, a list or numpy array of target energy values
    for each model.  This is primarily useful if you have already generated
    sticky ends, and want to choose ends that match (eg, if you chose DT ends
    and now want to choose TD ones).
    """
    if templates:
        templates = iter(templates)

    def endchooser(currentends, availends, energetics):
        nonlocal target_vals, templates

        if len(currentends) == 0 and not target_vals:
            # Starting out, we need to choose an initial end, and use that
            # to choose our target values.
            availfiltered = availends
            if templates:
                t = next(templates)
                for i, nt in enumerate(t):
                    if nt != 'n':
                        availfiltered = availfiltered[availfiltered[:, i] ==
                                                      lton[nt], :]
            dev = np.std(
                [en.matching_uniform(availends) for en in all_energetics],
                axis=0)
            choices = np.argsort(dev)
            choice = choices[np.random.randint(
                0, max(1, ceil(init_wigglefraction * len(availends))))]
            return availends[choice]
        else:
            if not target_vals:  # NOQA
                target_vals = [
                    en.matching_uniform(currentends[0:1])
                    for en in all_energetics
                ]
                LOGGER.debug("TVALS {}".format(target_vals))
            availfiltered = availends
            if templates:
                t = next(templates)
                LOGGER.debug("Template {}".format(t))
                for i, nt in enumerate(t):
                    if nt != 'n':
                        availfiltered = availfiltered[availfiltered[:, i] ==
                                                      lton[nt], :]
            if devmethod == 'dev':
                dev = np.sqrt(
                    np.sum(
                        np.array([
                            en.matching_uniform(availfiltered) - target_val
                            for en, target_val in zip(all_energetics, target_vals)
                        ])**2,
                        axis=0))
            elif devmethod == 'max':
                dev = np.max(np.abs(
                        np.array([
                            en.matching_uniform(availfiltered) - target_val
                            for en, target_val in zip(all_energetics, target_vals)
                        ])), axis=0)
            choices = np.argsort(dev)
            choice = choices[np.random.randint(
                0, max(1, ceil(len(choices) * next_wigglefraction)))]
            LOGGER.debug("Chose {}: {} from {}: {} / {}".format(
                availfiltered[choice:choice+1].tolist(),
                np.nonzero(choices == choice),
                len(availfiltered),
                dev[choice],
                np.concatenate([en.matching_uniform(availfiltered[choice:choice+1]) for en,
                 tv in zip(all_energetics, target_vals)])))
                         
            return availfiltered[choice]

    return endchooser


def deviation_score(all_ends, all_energetics, devmethod='dev'):

    if devmethod == 'dev':
        return np.sqrt(
            np.sum(
                np.var(
                    [
                        np.concatenate(
                            tuple(en.matching_uniform(ends) for ends in all_ends))
                        for en in all_energetics
                    ],
                    axis=1)))
    elif devmethod == 'max':
        return np.max([np.ptp(np.concatenate(
            tuple(en.matching_uniform(ends) for ends in all_ends)))
            for en in all_energetics])



import seaborn as sns
import pandas as pd
import xarray
import numpy as np

def _pandas_data(pairarrays, namelists, models, modelnames):
    ecarrays = [np.array([model.gse_all(pa.ends,pa.pairs,forcemulti=True) for model in models]) 
               for pa in pairarrays]
    eearrays = [np.array([model.gse_all(pa.ends,pa.ends,forcemulti=True) for model in models]) 
               for pa in pairarrays]
    ccarrays = [np.array([model.gse_all(pa.pairs,pa.pairs,forcemulti=True) for model in models]) 
               for pa in pairarrays]
    
    diags = pd.concat( tuple(pd.DataFrame(x.diagonal(axis1=1,axis2=2).T, columns=modelnames, index=n) for x, n
                            in zip(ecarrays, namelists) ) )
    ecm = [x.copy() for x in ecarrays]
    eem = [x.copy() for x in eearrays]
    ccm = [x.copy() for x in ccarrays]
    for x,y,z in zip(ecm,eem,ccm):
        x[np.tile(np.eye(x.shape[1],dtype=bool),(x.shape[0],1,1))] = np.nan
        y[(slice(0,None),)+np.tril_indices(y.shape[1],k=-1)] = np.nan
        z[(slice(0,None),)+np.tril_indices(y.shape[1],k=-1)] = np.nan
    nondiagsec_panels = [xarray.DataArray(x, [modelnames,n,[a+'/' for a in n]]).to_dataset('dim_0')
                         for x, n in zip(ecm, namelists)]
    nondiagsee_panels = [xarray.DataArray(x, [modelnames,n,n]).to_dataset('dim_0') 
                         for x, n in zip(eem, namelists)]
    nondiagscc_panels = [xarray.DataArray(x, [modelnames,[a+'/' for a in n],[a+'/' for a in n]]).to_dataset('dim_0')
                         for x, n in zip(ccm, namelists)]
    nondiag = pd.concat( tuple(x.to_dataframe() for x in nondiagsec_panels+nondiagsee_panels+nondiagscc_panels) )
    eca_panels = [xarray.DataArray(x, [modelnames,n,[a+'/' for a in n]]).to_dataset('dim_0') 
                         for x, n in zip(ecarrays, namelists)]
    eea_panels = [xarray.DataArray(x, [modelnames,n,n]).to_dataset('dim_0')
                         for x, n in zip(eearrays, namelists)]
    cca_panels = [xarray.DataArray(x, [modelnames,[a+'/' for a in n],[a+'/' for a in n]]).to_dataset('dim_0') 
                         for x, n in zip(ccarrays, namelists)]
    allv = pd.concat( tuple(x.to_dataframe() for x in eca_panels+eea_panels+cca_panels) )
    return (diags, nondiag, allv)



import numpy as np
from math import ceil


import logging
LOGGER = logging.getLogger(__name__)


class BasicChooser():
    def __init__(self, energetics, interaction, wiggle=0.0):
        self._energetics = energetics
        self._target = interaction
        self._wiggle = wiggle

    def choose(self, availends, currentends=None):
        ddiff = np.abs(self._energetics.gse(availends) - self._target)
        choices = np.flatnonzero(ddiff <= np.amin(ddiff)+self._wiggle)
        return availends[choices[np.random.randint(0, len(choices))]]



def deviation_score(all_ends, all_energetics, devmethod='dev'):

    if devmethod == 'dev':
        return np.sqrt(
            np.sum(
                np.var(
                    [
                        np.concatenate(
                            tuple(en.gse(ends) for ends in all_ends))
                        for en in all_energetics
                    ],
                    axis=1)))
    elif devmethod == 'max':
        return np.max([np.ptp(np.concatenate(
            tuple(en.gse(ends) for ends in all_ends)))
                       for en in all_energetics])


class MultiModelChooser:

    def __init__(self, all_energetics,
                 target_vals=None,
                 init_wigglefraction=1,
                 next_wigglefraction=0.1,
                 devmethod='dev'):
        self._tv = target_vals
        self._ae = all_energetics
        self._iwf = init_wigglefraction
        self._nwf = next_wigglefraction
        self._dm = devmethod

    def choose(self, availends, currentends):

        if (currentends is None or len(currentends) == 0) and not self._tv:
            # Starting out, we need to choose an initial end, and use that
            # to choose our target values.
            availfiltered = availends
            dev = np.std(
                [en.gse(availends) for en in self._ae],
                axis=0)
            choices = np.argsort(dev)
            choice = choices[np.random.randint(
                0, max(1, ceil(self._iwf * len(availends))))]
            return availends[choice]
        else:
            if not self._tv:  # NOQA
                self._tv = [
                    en.gse(currentends[0:1])
                    for en in self._ae
                ]
                LOGGER.debug("TVALS {}".format(self._tv))
            availfiltered = availends
            if self._dm == 'dev':
                dev = np.sqrt(
                    np.sum(
                        np.array([
                            en.gse(availfiltered) - target_val
                            for en, target_val in zip(self._ae, self._tv)
                        ])**2,
                        axis=0))
            elif self._dm == 'max':
                dev = np.max(np.abs(
                        np.array([
                            en.gse(availfiltered) - target_val
                            for en, target_val in zip(self._ae, self._tv)
                        ])), axis=0)
            choices = np.argsort(dev)
            choice = choices[np.random.randint(
                0, max(1, ceil(len(choices) * self._nwf)))]
            LOGGER.debug("Chose {}: {} from {}: {} / {}".format(
                availfiltered[choice:choice+1].tolist(),
                np.nonzero(choices == choice),
                len(availfiltered),
                dev[choice],
                np.concatenate([en.gse(availfiltered[choice:choice+1]) for en,
                                tv in zip(self._ae, self._tv)])))
                         
            return availfiltered[choice]






from __future__ import division
import numpy as np
import itertools
import logging

from .endclasses import endarray, lton, wc
from .energetics_basic import EnergeticsBasic

LOGGER = logging.getLogger(__name__)


def values_chunked(items, endtype, chunk_dim=10):
    """
    Given a list of lists of acceptable numbers for each position in a row of
    an array, create every possible row, and return an iterator that returns
    chunks of every possible row up to chunk_dim, iterating dimensions higher
    than chunk_dim.  This probably doesn't need to be called directly, and may
    have a _ added in the future.

    Return this as an endarray, with set endtype. This can be easily emoved
    for use elsewhere.
    """
    ilengths = [len(x) for x in items]
    n = len(items)
    items = [np.array(x) for x in items]
    if n > chunk_dim:
        p = n - chunk_dim
        q = chunk_dim
        outer = itertools.product(*(items[0:p]))
    else:
        p = 0
        q = n

        def outer_iter():
            yield ()

        outer = outer_iter()

    chunk = np.zeros(
        [np.prod(ilengths[p:]), len(items)], dtype=int).view(endarray)
    chunk.endtype = endtype
    chunk[:, p:] = np.indices(ilengths[p:]).reshape(q, -1).T
    for i in range(p, n):
        chunk[:, i] = items[i][chunk[:, i]]
    for seq in outer:
        chunk[:, :p] = seq
        yield chunk


def get_accept_set(endtype,
                   length,
                   interaction,
                   fdev,
                   maxendspurious,
                   spacefilter=None,
                   adjacents=['n', 'n'],
                   alphabet='n',
                   energetics=None):
    if not energetics:
        energetics = EnergeticsBasic()
    if not spacefilter:
        spacefilter = spacefilter_standard(interaction, interaction * fdev,
                                           maxendspurious)
    # Generate the template.
    if endtype == 'DT':
        template = [lton[adjacents[0]]] + [lton[alphabet.lower()]] \
                   * length + [lton[wc[adjacents[1]]]]
    elif endtype == 'TD':
        template = [lton[wc[adjacents[1]]]] + [lton[alphabet.lower()]] \
                   * length + [lton[adjacents[0]]]
    elif endtype == 'S':
        template = [lton[alphabet.lower()]]*length

    LOGGER.info("Length {0}, type {1}, adjacents {2}, alphabet {3}.".format(
        length, endtype, adjacents, alphabet))
    LOGGER.debug("Have template %s, endtype %s.", template, endtype)

    # Create the chunk iterator
    endchunk = values_chunked(template, endtype)

    # Use spacefilter to filter chunks down to usable sequences
    matcharrays = []
    chunknum = 0
    totchunks = None
    totends = np.product([len(x) for x in template])
    LOGGER.debug(
        "Have {0} ends in total before any filtering.".format(totends))
    for chunk in endchunk:
        matcharrays.append(spacefilter(chunk, energetics))
        if not totchunks:
            totchunks = totends // len(chunk)
        chunknum += 1
        LOGGER.debug("Found {0} filtered ends in chunk {1} of {2}.".format(
            len(matcharrays[-1]), chunknum, totchunks))
    LOGGER.debug("Done with spacefiltering.")
    availends = np.vstack(matcharrays).view(endarray)
    availends.endtype = endtype
    return availends


def _make_avail(endtype,
                length,
                spacefilter,
                endfilter,
                endchooser,
                energetics,
                adjacents=['n', 'n'],
                num=0,
                numtries=1,
                oldendfilter=None,
                oldends=[],
                alphabet='n'):
        # Generate the template.
    if endtype == 'DT':
        template = [lton[adjacents[0]]] + [lton[alphabet.lower()]] \
                   * length + [lton[wc[adjacents[1]]]]
    elif endtype == 'TD':
        template = [lton[wc[adjacents[1]]]] + [lton[alphabet.lower()]] \
                   * length + [lton[adjacents[0]]]
    elif endtype == 'S':
        template = [lton[alphabet.lower()]]*length
    
    LOGGER.info("Length {0}, type {1}, adjacents {2}, alphabet {3}.".format(
        length, endtype, adjacents, alphabet))
    LOGGER.debug("Have template %s, endtype %s", template, endtype)

    # Create the chunk iterator
    endchunk = values_chunked(template, endtype)

    # Use spacefilter to filter chunks down to usable sequences
    matcharrays = []
    chunknum = 0
    totchunks = None
    totends = np.product([len(x) for x in template])
    LOGGER.debug(
        "Have {0} ends in total before any filtering.".format(totends))
    for chunk in endchunk:
        matcharrays.append(spacefilter(chunk, energetics))
        if not totchunks:
            totchunks = totends // len(chunk)
        chunknum += 1
        LOGGER.debug("Found {0} filtered ends in chunk {1} of {2}.".format(
            len(matcharrays[-1]), chunknum, totchunks))
    LOGGER.debug("Done with spacefiltering.")
    availends = np.vstack(matcharrays).view(endarray)
    availends.endtype = endtype

    # Use endfilter to filter available sequences taking into account old
    # sequences.
    if len(oldends) > 0:
        if oldendfilter:
            availends = oldendfilter(oldends, None, availends, energetics)
        else:
            availends = endfilter(oldends, None, availends, energetics)

    return availends


def find_end_set_uniform(endtype,
                         length,
                         spacefilter,
                         endfilter,
                         endchooser,
                         energetics,
                         adjacents=['n', 'n'],
                         num=0,
                         numtries=1,
                         oldendfilter=None,
                         oldends=[],
                         alphabet='n',
                         _presetavail=False):
    """
    Find a set of ends of uniform length and type satisfying uniform
    constraint functions (eg, constrant functions are the same for each
    end).

    This function is intended to be complicated and featureful. If you want
    something simpler, try easy_ends

    Parameters
    ----------

    endtype : str
      right now 'DT' for 3'-terminal ends, and 'TD' for
      5'-terminal ends,
    length : int
      length of ends, not including adjacent bases, if applicable.
    adjacents : list of str
      (defaults to ['n','n']): acceptable bases for adjacents
      (eg, ['n','n'] or ['c', 'c']) for the ends and their complements,
    num : int
      (defaults to 0): number of ends to find (0 keeps finding until
      available ends are exhausted)
    numtries : int
      (defaults to 1): if > 1, the function will return a list of
      sets of ends that all individually satisfy the constraints, so that
      the best one can be selected manually
    spacefilter: function
        a "spacefilter" function that takes endarrays and
        filters them down to ends that, not considering spurious
        interactions, are acceptable.
    endfilter: function
        an "endfilter" function that takes current ends in the
        set, available ends (filtered with current ends), and new ends added,
        and filters the available ends, considering interactions between ends
        (eg, spurious interactions).
    endchooser : function
        an "endchooser" function that takes current ends in the
        set and available ends, and returns a new end to add to the set.
    energetics : function
        an "energyfunctions" class that provides the energy
        functions for everything to use.
    oldends : endarray
        an endarray of old ends to consider as part of the set
    alphabet : str
        a single letter specifying what the alphabet for the ends
        should be (eg, four or three-letter code)
    oldendfilter : str
        a different "endfilter" function for use when filtering
        the available ends using interactions with old ends. This is normally
        not useful, but can be useful if you want, for example, to create a
        sets with higher cross-interactions between two subsets than within
        the two subsets.
    

    Returns
    -------

    endarray
      an endarray of generated ends, including provided old ends
    """

    if len(oldends) > 0:
        if type(oldends[0]) is str:
            oldends = endarray(oldends, endtype)
    
    if isinstance(_presetavail, endarray):
        startavail = _presetavail
    else:
        startavail = _make_avail(endtype,
                                 length,
                                 spacefilter,
                                 endfilter,
                                 endchooser,
                                 energetics,
                                 adjacents,
                                 num,
                                 numtries,
                                 oldendfilter,
                                 oldends,
                                 alphabet)
    endsets = []
    availends = startavail.copy()
    LOGGER.debug("Starting with {0} ends.".format(len(availends)))
    while len(endsets) < numtries:
        curends = oldends
        availends = startavail.copy()
        numends = 0
        while True:
            newend = endarray(
                np.array([endchooser(curends, availends, energetics)]),
                endtype)
            LOGGER.debug("Chose end {0}.".format(repr(newend)))
            newend.endtype = endtype
            availends = endfilter(newend, curends, availends, energetics)
            LOGGER.debug("Done filtering.")
            if curends is None:
                curends = newend
            elif len(curends) == 0:
                curends = newend
            else:
                curends = curends.append(newend)
            numends += 1
            LOGGER.debug("Now have {0} ends in set, and {1} ends available.".format(numends, len(availends)))
            if len(availends) == 0:
                LOGGER.info("Found {0} ends.".format(numends))
                break
            if numends >= num and num > 0:
                break
        endsets.append(curends)

    # Verification:
    # Note: this currently gives weird output that is not helpful when it fails.
    # But if this fails, you've done something very weird, most likely, because
    # this is just internal sanity checking.
    for endset in endsets:
        oldr = np.arange(0, len(oldends))
        newr = np.arange(len(oldends), len(endset))
        allr = np.arange(0, len(endset))
        # All new ends must satisfy old ends:
        if oldendfilter is None and len(oldends) > 0:
            assert np.asarray(
                endfilter(endset[oldr, :], None,
                          endset[newr, :], energetics) ==
                endset[newr, :]).all()
        elif len(oldends) > 0:
            assert np.asarray(
                oldendfilter(endset[oldr, :], None,
                             endset[newr, :], energetics) ==
                endset[newr, :]).all()
        # Each new end must allow all others
        for i in newr:
            if oldendfilter is None:
                assert np.asarray(
                    endfilter(endset[i, :][None, :], None,
                              endset, energetics) ==
                    endset[i != allr, :]).all()
            else:
                assert np.asarray(
                    oldendfilter(endset[i, :][None, :], None,
                                 endset[oldr, :], energetics) ==
                    endset[oldr, :]).all()
                assert np.asarray(
                    endfilter(endset[i, :][None, :], None,
                              endset[newr, :], energetics) ==
                    endset[newr[i != newr], :]).all()

    if len(endsets) > 1:
        return endsets
    else:
        if _presetavail is None or isinstance(_presetavail,endarray):
            return endsets[0], startavail
        else:
            return endsets[0]


def enhist(endtype,
           length,
           adjacents=['n', 'n'],
           alphabet='n',
           bins=None,
           energetics=None,
           plot=False,
           color='b'):
    if endtype == 'DT':
        template = [lton[adjacents[0]]] +\
                   [lton[alphabet.lower()]] * length + [lton[wc[adjacents[1]]]]
    elif endtype == 'TD':
        template = [lton[wc[adjacents[1]]]] +\
                   [lton[alphabet.lower()]] * length + [lton[adjacents[0]]]
    elif endtype == 'S':
        template = [lton[alphabet.lower()]]*length
    
    if not energetics:
        energetics = EnergeticsBasic()

    minbin = 0.8 * energetics.matching_uniform(
        endarray([([0, 3] * length)[0:length + 2]], endtype))
    maxbin = 1.1 * energetics.matching_uniform(
        endarray([([1, 2] * length)[0:length + 2]], endtype))

    if not bins:
        bins = np.arange(minbin, maxbin, 0.1)

    LOGGER.debug("Have template {0} and type {1}.".format(template, endtype))

    # Create the chunk iterator
    endchunk = values_chunked(template, endtype)

    hist = np.zeros(len(bins) - 1, dtype='int')
    totends = np.product([len(x) for x in template])
    finishedends = 0
    info = {'min': np.inf, 'max': -np.inf, 'mean': 0}
    for chunk in endchunk:
        matchens = energetics.matching_uniform(chunk)
        hist += np.histogram(matchens, bins)[0]
        info['max'] = max(info['max'], np.amax(matchens))
        info['min'] = min(info['min'], np.amin(matchens))
        info['mean'] = (info['mean']*(finishedends)/(len(chunk)+finishedends)
                        + np.mean(matchens) * len(chunk) /
                        (len(chunk)+finishedends))
        finishedends += len(matchens)
        LOGGER.debug("Done with {0}/{1} ends.".format(finishedends, totends))

    x = (bins[:-1] + bins[1:]) / 2
    n = hist
    info['emean'] = np.sum(n * x, dtype='double') / np.sum(n, dtype='int64')
    info['estd'] = np.sqrt(
        np.sum(n * (
            x - info['emean'])**2, dtype='double') / np.sum(n, dtype='int64'))
    cs = np.cumsum(n)
    info['emedian'] = x[np.flatnonzero(cs >= cs[-1] / 2.0)[0]]

    if plot:
        import matplotlib.pyplot as plt

        plt.bar(
            bins[:-1],
            hist,
            width=(bins[1] - bins[0]),
            label="Type {3}, Length {0}, Adjs {1}, Alph {2}".format(
                length, adjacents, alphabet, endtype),
            color=color)
        plt.title(
            "Matching Energies of Ends of Type {3}, Length {0}, Adjs {1}, Alph {2}".
            format(length, adjacents, alphabet, endtype))
        plt.xlabel("Standard Free Energy (-kcal/mol)")
        plt.ylabel("Number of Ends")
        # plt.show()

    return (hist, bins, info)


def easyends(endtype,
             endlength,
             number=0,
             interaction=None,
             fdev=0.05,
             maxspurious=0.5,
             maxendspurious=None,
             tries=1,
             oldends=[],
             adjs=['n', 'n'],
             energetics=None,
             alphabet='n',
             echoose=None,
             absolute=False,
             _presetavail=False):
    """
    Easyends is an attempt at creating an easy-to-use function for finding sets
    of ends.

    * endtype: specifies the type of end being considered. The system for
      classifying end types goes from 5' to 3', and consists of letters
      describing each side of the end. For example, an end that starts after a
      double-stranded region on the 5' side and ends at the end of the strand
      would be 'DT', while one that starts at the beginning of a strand on the
      5' side and ends in a double-stranded region would be 'TD'. 'T' stands
      for terminal, 'D' stands for double-stranded region, and 'S' stands for
      single-stranded region. 'S', however, is not currently supported.
    * endlength: specifies the length of end being considered, not including
      adjacent bases.
    * number (optional): specifies the number of ends to find.  If zero or not
      provided, easyends tries to find as many ends as possible.
    * interaction (optional): a positive number corresponding to the desired
      standard free energy for hybridization of matching sticky ends. If not
      provided, easyends calculates an optimal value based on the sequence
      space.
    * fdev (default 0.05): the fractional deviation (above or below) of
      allowable matching energies.  maxspurious (default 0.5): the maximum
      spurious interaction, as a fraction of the matching interaction.
    * maxendspurious (default None): if provided, maxspurious is only used for
      spurious interactions between ends defined as ends, and ends defined as
      complements. Maxendspurious is then the maximum spurious interaction
      between ends and ends, and complements and complements. In a system
      where spurious interactions between ends and complements are more important
      than other spurious interactions, this can allow for better sets of ends.
    * tries (default 1): if > 1, easyends will return a list of sets of ends,
      all satisfying the constraints.
    * oldends (optional): a list of ends to be considered as already part of
      the set.
    * adjacents (default ['n','n']): allowable adjacent bases for ends and
      complements.
    * absolute (default False): fdev, maxspurious, and maxendspurious to be
      interpreted as absolute kcal/mol values rather than fractional values.
    * energetics (optional): an energetics class providing the energy
      calculation functions. You probably don't need to change this.
    * alphabet (default 'n'): The alphabet to use for ends, allowing
      for three-letter codes.
    """

    if not energetics:
        efunc = EnergeticsBasic()
    else:
        efunc = energetics
    if (not interaction) or (interaction == 0):
        interaction = enhist(
            endtype,
            endlength,
            energetics=efunc,
            adjacents=adjs,
            alphabet=alphabet)[2]['emedian']
        LOGGER.info("Calculated optimal interaction energy is {0}.".format(
            interaction))
    if not absolute:
        mult = interaction
    else:
        mult = 1.0
        
    maxcompspurious = maxspurious * mult
    if not maxendspurious:
        maxendspurious = maxspurious * mult
    else:
        maxendspurious = maxendspurious * mult

    sfilt = spacefilter_standard(interaction, mult * fdev,
                                 maxendspurious)
    efilt = endfilter_standard_advanced(maxcompspurious, maxendspurious)
    if not echoose:
        echoose = endchooser_standard(interaction)

    return find_end_set_uniform(
        endtype,
        endlength,
        sfilt,
        efilt,
        echoose,
        energetics=efunc,
        numtries=tries,
        oldends=oldends,
        adjacents=adjs,
        num=number,
        alphabet=alphabet,
        _presetavail=_presetavail)


def easy_space(endtype,
               endlength,
               interaction=None,
               fdev=0.05,
               maxspurious=0.5,
               maxendspurious=None,
               tries=1,
               oldends=[],
               adjs=['n', 'n'],
               energetics=None,
               alphabet='n',
               echoose=None):
    length = endlength
    if not energetics:
        efunc = EnergeticsBasic()
        energetics = efunc
    else:
        efunc = energetics
    if (not interaction) or (interaction == 0):
        interaction = enhist(
            endtype,
            endlength,
            energetics=efunc,
            adjacents=adjs,
            alphabet=alphabet)[2]['emedian']
        LOGGER.info("Calculated optimal interaction energy is {0}.".format(
            interaction))
    maxcompspurious = maxspurious * interaction
    if not maxendspurious:
        maxendspurious = maxspurious * interaction
    else:
        maxendspurious = maxendspurious * interaction

    sfilt = spacefilter_standard(interaction, interaction * fdev,
                                 maxendspurious)
    spacefilter = sfilt

    if not echoose:
        echoose = endchooser_standard(interaction)

    adjacents = adjs

    if endtype == 'DT':
        template = [lton[adjacents[0]]] + [lton[alphabet.lower()]] \
                   * length + [lton[wc[adjacents[1]]]]
    elif endtype == 'TD':
        template = [lton[wc[adjacents[1]]]] + [lton[alphabet.lower()]] \
                   * length + [lton[adjacents[0]]]

    # Create the chunk iterator
    endchunk = values_chunked(template, endtype)

    # Use spacefilter to filter chunks down to usable sequences
    matcharrays = []
    chunknum = 0
    totchunks = None
    totends = np.product([len(x) for x in template])
    LOGGER.info(
        "Have {0} ends in total before any filtering.".format(totends))
    for chunk in endchunk:
        matcharrays.append(spacefilter(chunk, energetics))
        if not totchunks:
            totchunks = totends // len(chunk)
        chunknum += 1
        LOGGER.debug("Found {0} filtered ends in chunk {1} of {2}.".format(
            len(matcharrays[-1]), chunknum, totchunks))
    LOGGER.debug("Done with spacefiltering.")
    availends = np.vstack(matcharrays).view(endarray)
    availends.endtype = endtype

    availendsr = np.repeat(availends, len(availends), axis=0)
    availendst = np.tile(availends, (len(availends), 1))

    vals_ee = energetics.uniform(availendsr.ends, availendst.ends)
    vals_ec = energetics.uniform(availendsr.ends, availendst.comps)
    vals_ce = energetics.uniform(availendsr.comps, availendst.ends)
    vals_cc = energetics.uniform(availendsr.comps, availendst.comps)

    vals_tf = ((vals_ee < maxendspurious) & (vals_cc < maxendspurious) &
               (vals_ec < maxcompspurious) & (vals_ce < maxcompspurious))
    zipendsnf = zip(availendsr.tolist(), availendst.tolist())
    zipends = [zipendsnf[x] for x in np.flatnonzero(vals_tf)]

    return zipends


def spacefilter_standard(desint, dev, maxself):
    """
    A spacefilter function: filters to ends that have a end-complement
    interaction of between desint-dev and desint+dev, and a self-interaction
    (end-end or comp-comp) of less than maxself.
    """

    def spacefilter(fullends, energetics):
        matchenergies = energetics.matching_uniform(fullends)
        g4 = np.zeros(fullends.shape[0])
        for w in range(0, (fullends.shape[1] - 3)):
            g4 += (np.sum(
                np.array(fullends[:, w:(w + 4)] == [2, 2, 2, 2]), axis=1) == 4)
            g4 += (np.sum(
                np.array(fullends[:, w:(w + 4)] == [1, 1, 1, 1]), axis=1) == 4)
        i = np.flatnonzero((matchenergies < desint + dev) &
                           (matchenergies > desint - dev) & (g4 == 0))
        matchenergies = matchenergies[i]
        fullends = fullends[i]
        selfselfenergies = energetics.uniform(fullends.ends, fullends.ends)
        compcompenergies = energetics.uniform(fullends.comps, fullends.comps)
        i = np.flatnonzero((selfselfenergies < maxself) & (compcompenergies <
                                                           maxself))
        return fullends[i]

    return spacefilter


def endfilter_standard(maxspurious):
    """
    An endfilter function: filters out ends that have any (end-end, end-comp,
    comp-end, comp-comp) interactions with new ends above maxspurious.
    """

    def endfilter(newends, currentends, availends, energetics):
        endendspurious = energetics.uniform(
            np.repeat(newends.ends, len(availends), 0),
            np.tile(availends.ends, (len(newends), 1))).reshape(
                len(availends), len(newends), order='F')
        endcompspurious = energetics.uniform(
            np.repeat(newends.ends, len(availends), 0),
            np.tile(availends.comps, (len(newends), 1))).reshape(
                len(availends), len(newends), order='F')
        compendspurious = energetics.uniform(
            np.repeat(newends.comps, len(availends), 0),
            np.tile(availends.ends, (len(newends), 1))).reshape(
                len(availends), len(newends), order='F')
        compcompspurious = energetics.uniform(
            np.repeat(newends.comps, len(availends), 0),
            np.tile(availends.comps, (len(newends), 1))).reshape(
                len(availends), len(newends), order='F')
        highspurious = np.amax(
            np.hstack((endendspurious, compendspurious, endcompspurious,
                       compcompspurious)), 1)

        return availends[highspurious < maxspurious]

    return endfilter


def endfilter_standard_advanced(maxcompspurious, maxendspurious):
    """
    An endfilter function: filters out ends that have end-comp or comp-end
    interactions above maxcompspurious, and end-end or comp-comp interactions
    above maxendspurious.
    """

    def endfilter(newends, currentends, availends, energetics):
        endendspurious = energetics.uniform(
            np.repeat(newends.ends, len(availends), 0),
            np.tile(availends.ends, (len(newends), 1))).reshape(
                len(availends), len(newends), order='F')
        endcompspurious = energetics.uniform(
            np.repeat(newends.ends, len(availends), 0),
            np.tile(availends.comps, (len(newends), 1))).reshape(
                len(availends), len(newends), order='F')
        compendspurious = energetics.uniform(
            np.repeat(newends.comps, len(availends), 0),
            np.tile(availends.ends, (len(newends), 1))).reshape(
                len(availends), len(newends), order='F')
        compcompspurious = energetics.uniform(
            np.repeat(newends.comps, len(availends), 0),
            np.tile(availends.comps, (len(newends), 1))).reshape(
                len(availends), len(newends), order='F')
        highendspurious = np.amax(
            np.hstack((endendspurious, compcompspurious)), 1)
        highcompspurious = np.amax(
            np.hstack((compendspurious, endcompspurious)), 1)

        return availends[(highendspurious < maxendspurious)
                         & (highcompspurious < maxcompspurious)]

    return endfilter


def energy_array_uniform(seqs, energetics):
    """
    Given an endarray and a set of sequences, return an array of the
    interactions between them, including their complements.
    """
    seqs = seqs.ends.append(seqs.comps)
    return energetics.uniform(
        np.repeat(seqs, seqs.shape[0], 0), np.tile(
            seqs, (seqs.shape[0], 1))).reshape((seqs.shape[0], seqs.shape[0]))


def endchooser_standard(desint, wiggle=0.0):
    """
    An endchooser function: return a random end with end-comp energy closest to
    desint.
    """

    def endchooser(currentends, availends, energetics):
        ddiff = np.abs(energetics.matching_uniform(availends) - desint)
        choices = np.flatnonzero(ddiff <= np.amin(ddiff)+wiggle)
        newend = availends[choices[np.random.randint(0, len(choices))]]
        return newend

    return endchooser


def endchooser_random():
    """
    An endchooser function: return a random end with end-comp energy closest to
    desint.
    """

    def endchooser(currentends, availends, energetics):
        newend = availends[np.random.randint(0, len(availends))]
        return newend

    return endchooser






from .seqclasses import SeqPairArray
import numpy as np


class SeqFilter(object):

    def __init__(self, energymodel, targetint,
                 maxdeviation=0.05, maxspurious=0.5,
                 maxendendspurious=None,
                 maxtype='fractional'):
        """
Create a SeqFilter instance.

Parameters
----------
energymodel : EnergyModel
    the energy model to use.  This can be changed later with
    the energymodel attribute.

targetint : float
    the target interaction (usually in kcal/mol).  This can be
    changed later with the targetint attribute.

maxdeviation : float, optional
    the maximum deviation allowed for an end from the target
    interaction.  If maxtype = 'fractional', then this is interpreted
    as a fraction of targetint above or below targetint. If maxtype =
    'relative' or 'absolute', this is interpreted as a kcal/mol value
    above or below targetint.
    (default 0.05)

maxspurious : float, optional
    the maximum spurious interaction allowed between an end and other
    ends / complements.  If maxtype = 'fractional', then this is
    interpreted as a fraction of targetint.  If maxtype = 'absolute',
    this is interpreted as an absolute kcal/mol threshold.  If
    maxtype = 'relative', this is interpreted as a kcal/mol amount
    below targetint.
    (default 0.5)

maxendendspurious : float, optional
    if None, then maxspurious will be used for everything.  If specified,
    then maxspurious will only apply to interactions between ends and
    the complements of other ends, while maxendendspurious will apply to
    interactions between ends and other ends.  This could be useful if,
    for example, your system design is such that ends can't easily interact
    with other ends.
    (default None)

maxtype : {'relative', 'absolute', 'fractional'}, optional
    determines how the max parameters are to be interpreted.  If
    'fractional', each is considered as a fraction of the targetint.
    If 'relative' or 'absolute', all are considered to be in kcal/mol.
    (default 'fractional')
        """
        self.energymodel = energymodel
        self._targetint = targetint
        self._maxdeviation = maxdeviation
        self._maxspurious = maxspurious
        self._maxendendspurious = maxendendspurious
        self._maxtype = maxtype
        self._recalcparams()

    def _recalcparams(self):
        """
        Recalculate parameters (maxdeviation, maxspurious, 
        maxendendspurious) on the basis of maxtype and targetint.

        This is called on initialization and whenever any parameter
        is changed.  You should never need to call it manually.
        """

        if self._maxendendspurious is None:
            mees = self._maxspurious
        
        if self._maxtype == 'fractional':
            self._rmd = self._maxdeviation * self._targetint
            self._rms = self._maxspurious * self._targetint
            self._rmees = mees * self._targetint
        elif self._maxtype == 'absolute':
            self._rmd = self._maxdeviation
            self._rms = self._maxspurious
            self._rmees = mees
        elif self._maxtype == 'relative':
            self._rmd = self._maxdeviation
            self._rms = self._targetint - self._maxspurious
            self._rmees = self._targetint - mees
        else:
            raise ValueError("{} not a recognized maxtype".format(
                self._maxtype))

    def filterspace(self, seqpairs):
        """
Filter a SeqPairArray down to only seqpairs that could satisfy the
filter's parameters: ensure that matching (seq-pair) interactions are
within the maxdeviation, and self (seq-seq, pair-pair) interactions are
within maxspurious or maxendendspurious.
        
Parameters
----------

seqpairs : SeqPairArray
    the seqpairs to filter.

Returns
-------
out : SeqPairArray
    the filtered seqpairs
        """
        ma = self.energymodel.gse(seqpairs)
        ss = self.energymodel.gse_all(seqpairs.seqs, seqpairs.seqs)
        cc = self.energymodel.gse_all(seqpairs.comps, seqpairs.comps)
        filt = ((np.abs(ma-self._targetint) < self._rmd) &
                (ss < self._rmees) & (cc < self._rmees))
        f1 = seqpairs[filt]

        # now filter poly-G:
        g4 = np.zeros(f1.shape[0])
        for w in range(0, (f1.shape[1] - 3)):
            g4 += (np.sum(
                np.array(f1[:, w:(w + 4)] == [2, 2, 2, 2]), axis=1) == 4)
            g4 += (np.sum(
                np.array(f1[:, w:(w + 4)] == [1, 1, 1, 1]), axis=1) == 4)
        return f1[g4 == 0]

    def filterseqs(self, avail, new):
        """
Filter a SeqPairArray, given the addition of new seqs: filter avail
by the interactions each seqpair has with the seqpair/seqpairs in new.

Parameters
----------

avail : SeqPairArray
    The seqpairs to be filtered.

new : SeqPairArray
    New seqs (seqs that have not yet been used to filter 
    available ends) that should be used to filter avail.

Returns
-------
out : SeqPairArray
    The filtered seqpairs that are still available.
        """

        ss = self.energymodel.gse_all(new.seqs, avail.seqs, forcemulti=True)
        sc = self.energymodel.gse_all(new.seqs, avail.comps, forcemulti=True)
        cs = self.energymodel.gse_all(new.comps, avail.seqs, forcemulti=True)
        cc = self.energymodel.gse_all(new.comps, avail.comps, forcemulti=True)

        filt = ((ss < self._rmees) &
                (cc < self._rmees) &
                (sc < self._rms) &
                (cs < self._rms)).all(axis=0)

        return avail[filt]



from stickydesign import energetics_daoe
from stickydesign import energetics_basic_old
from stickydesign import energetics_basic
from stickydesign import endarray
import pytest
import numpy as np

@pytest.fixture()
def sets_daoe():
        r5dt = endarray(np.random.randint(low=0, high=4, size=(100, 7)), 'DT')
        r5td = endarray(np.random.randint(low=0, high=4, size=(100, 7)), 'TD')
        r10dt = endarray(
            np.random.randint(low=0, high=4, size=(100, 12)), 'DT')
        r10td = endarray(
            np.random.randint(low=0, high=4, size=(100, 12)), 'TD')
        return [r5dt, r5td, r10dt, r10td]

@pytest.fixture()
def sets_basic():
    r7s = endarray(np.random.randint(low=0, high=4, size=(100, 7)), 'S')
    r12s = endarray(np.random.randint(low=0, high=4, size=(100, 12)), 'S')
    return [r7s, r12s]

@pytest.fixture()
def en_daoe():
    return energetics_daoe.EnergeticsDAOE()

@pytest.fixture()
def en_basicold():
    return energetics_basic_old.EnergeticsBasicOld()

@pytest.fixture()
def en_basic():
    return energetics_basic.EnergeticsBasic()


def test_daoe_matching_energies_match(sets_daoe, en_daoe):
    for s in sets_daoe:
        r1 = en_daoe.matching_uniform(s)
        r2 = en_daoe.uniform(s.ends, s.comps)
        print(repr(s))
        np.testing.assert_array_almost_equal(r1, r2)

def test_daoe_symmetry(sets_daoe, en_daoe):
    for s in sets_daoe:
        r1 = en_daoe.uniform(s[:50], s[50:])
        r2 = en_daoe.uniform(s[50:], s[:50])
        np.testing.assert_array_almost_equal(r1, r2)

def test_basic_matching_energies_match(sets_basic, en_basic):
    for s in sets_basic:
        r1 = en_basic.matching_uniform(s)
        r2 = en_basic.uniform(s.ends, s.comps)
        print(repr(s))
        np.testing.assert_array_almost_equal(r1, r2)

def test_basic_symmetry(sets_basic, en_basic):
    for s in sets_basic:
        r1 = en_basic.uniform(s[:50], s[50:])
        r2 = en_basic.uniform(s[50:], s[:50])
        np.testing.assert_array_almost_equal(r1, r2)

def test_basic_old_matching_energies_match(sets_daoe, en_basicold):
    for s in sets_daoe:
        r1 = en_basicold.matching_uniform(s)
        r2 = en_basicold.uniform_loopmismatch(s.ends, s.comps)
        r3 = en_basicold.uniform_danglemismatch(s.ends, s.comps)
        print(repr(s))
        np.testing.assert_array_almost_equal(r1, r2)
        np.testing.assert_array_almost_equal(r1, r3)

def test_basic_old_symmetry(sets_daoe, en_basicold):
    for s in sets_daoe:
        r1 = en_basicold.uniform(s[:50], s[50:])
        r2 = en_basicold.uniform(s[50:], s[:50])
        np.testing.assert_array_almost_equal(r1, r2)
